// Ansteuerung eines HD44780 kompatiblen LCD im 4-Bit-Interfacemodus
// http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial
//
// Configure pins in the header-file

#include <avr/io.h>
#include "lcd-routines.h"
#include <util/delay.h>

volatile uint8_t selectedController;

// sent data to the LCD
void lcd_data(uint8_t temp1)
{
    unsigned char temp2 = temp1;

    LCD_PORT |= (1<<LCD_RS);        // RS auf 1 setzen

    temp1 = temp1 >> 4;
    temp1 = temp1 & 0x0F;
    LCD_PORT &= 0xF0;
    LCD_PORT |= temp1;               // setzen
    lcd_enpulse(selectedController);

    temp2 = temp2 & 0x0F;
    LCD_PORT &= 0xF0;
    LCD_PORT |= temp2;               // setzen
    lcd_enpulse(selectedController);

    _delay_us(42);
    //lcd_waitbusy();
}

// sent a command to the LCD
void lcd_command(uint8_t temp1, uint8_t controller)
{
    unsigned char temp2 = temp1;

    LCD_PORT &= ~(1<<LCD_RS);        // RS auf 0 setzen

    temp1 = temp1 >> 4;              // oberes Nibble holen
    temp1 = temp1 & 0x0F;            // maskieren
    LCD_PORT &= 0xF0;
    LCD_PORT |= temp1;               // setzen
    lcd_enpulse(controller);

    temp2 = temp2 & 0x0F;            // unteres Nibble holen und maskieren
    LCD_PORT &= 0xF0;
    LCD_PORT |= temp2;               // setzen
    lcd_enpulse(controller);

    _delay_us(43);
    //lcd_waitbusy();
}

// Creates a pulse at the enable pins
// controller 0: both controllers
//            1: controller 1
//            2: controller 2
void lcd_enpulse(uint8_t controller)
{
    if(controller == 1)
    {
        LCD_PORT |= (1<<LCD_EN1);
        _delay_us(1);
        LCD_PORT &= ~(1<<LCD_EN1);
    }
    else if (controller == 2)
    {
        LCD_PORT |= (1<<LCD_EN2);
        _delay_us(1);
        LCD_PORT &= ~(1<<LCD_EN2);
    }
    else
    {
        LCD_PORT |= (1<<LCD_EN1);
        LCD_PORT |= (1<<LCD_EN2);
        _delay_us(1);
        LCD_PORT &= ~(1<<LCD_EN1);
        LCD_PORT &= ~(1<<LCD_EN2);
    }
}


// Initialisation:
// execute at the very beginning of the program.
void lcd_init(void)
{
    // Port auf Ausgang schalten
    LCD_DDR = (1<<LCD_DB4) | (1<<LCD_DB5) | (1<<LCD_DB6) | (1<<LCD_DB7) | (1<<LCD_RS) | (1<<LCD_EN1) | (1<<LCD_EN2) | (1<<LCD_RW);

    // muss 3mal hintereinander gesendet werden zur Initialisierung
    _delay_ms(15);
    LCD_PORT &= 0xF0;
    LCD_PORT |= 0x03;
    LCD_PORT &= ~(1<<LCD_RS);      // RS auf 0
    lcd_enpulse(0);
    _delay_ms(5);
    lcd_enpulse(0);
    _delay_ms(1);
    lcd_enpulse(0);
    _delay_ms(1);

    // Activate 4-bit mode
    LCD_PORT &= 0xF0;
    LCD_PORT |= 0x02;
    lcd_enpulse(0);
    _delay_ms(1);

    // 4Bit / 2 Lines / 5x7 Font
    lcd_command(0x28, 0);

    // Display on / Cursor off / no flashing cursor
    lcd_command(0x0C, 0);

    // inkrement / no scrolling
    lcd_command(0x06, 0);

    selectedController = 1;

    lcd_clear();
}

// sent command to clear the display
void lcd_clear(void)
{
    lcd_command(CLEAR_DISPLAY, 0);
    _delay_ms(5);
}

// sent command to set the cursor at the home position
void lcd_home(void)
{
    lcd_command(CURSOR_HOME, 0);
    selectedController = 1;
    _delay_ms(5);
}

// set the cursor at line y (0 ... 3) and column x (0 ...39)
void set_cursor(uint8_t x, uint8_t y)
{
    uint8_t tmp;

    switch (y)
    {
    case 0:     // Controller 1, first line
        selectedController = 1;
        tmp=0x80+x;
        break;
    case 1:     // Controller 1, second line
        selectedController = 1;
        tmp=0xC0+x;
        break;
    case 2:     // Controller 2, first line
        selectedController = 2;
        tmp=0x80+x;
        break;
    case 3:     // Controller 2, second line
        selectedController = 2;
        tmp=0xC0+x;
        break;
    default:
        return;
    }
    lcd_command(tmp, selectedController);
}

// writes a string to the LCD
void lcd_string(char *data)
{
    while (*data)
    {
        lcd_data(*data);
        data++;
    }
}

// rs 1: read Data          controller 1: Read controller 1
//    0: read Busyflag                 2: Read controller 2
uint8_t lcd_read(uint8_t rs, uint8_t controller)
{
    uint8_t data;

    // Set/clear RS bit
    if(rs)
        LCD_PORT |= (1<<LCD_RS);
    else
        LCD_PORT &= ~(1<<LCD_RS);

    LCD_DDR &= ~( (1<<LCD_DB4) | (1<<LCD_DB5) | (1<<LCD_DB6) | (1<<LCD_DB7) );  // config data-pins as input


    if(controller == 1)
    {
        LCD_PORT |= (1<<LCD_EN1);
        _delay_us(1);
        data = LCD_PORT;
        data = data << 4;
        LCD_PORT &= ~(1<<LCD_EN1);

        LCD_PORT |= (1<<LCD_EN1);
        _delay_us(1);
        data |= LCD_PORT & 0x0F;
        LCD_PORT &= ~(1<<LCD_EN1);
    }
    else
    {

    }

    return data;
}

// Wartet solange das Busyflag vom LCD gesetzt ist
void lcd_waitbusy(void)
{
    LCD_PORT |= (1<<LCD_RW);
    LCD_PORT &= ~(1<<LCD_RS);
    LCD_DDR &= ~(1<<LCD_DB7);   // DB7 as input

    while(LCD_PORT & (1<<LCD_DB7))
    {
        __asm__ __volatile__( "nop" );
    }

    LCD_PORT &= ~(1<<LCD_RW);
    LCD_PORT &= ~(1<<LCD_RS);
    LCD_DDR |= (1<<LCD_DB7);   // DB7 as output
}
